@@ -1,4 +1,4 @@
-# Copyright 2012-2018 The Meson development team
+# Copyright 2012-2019 The Meson development team
 
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -17,12 +17,15 @@ import pickle, os, uuid, shlex
 import sys
 from pathlib import PurePath
 from collections import OrderedDict
-from .mesonlib import MesonException
-from .mesonlib import default_libdir, default_libexecdir, default_prefix
+from .mesonlib import (
+    MesonException, default_libdir, default_libexecdir, default_prefix
+)
+from .wrap import WrapMode
 import ast
 import argparse
+import configparser
 
-version = '0.47.1'
+version = '0.49.1'
 backendlist = ['ninja', 'vs', 'vs2010', 'vs2015', 'vs2017', 'xcode']
 
 default_yielding = False
@@ -138,9 +141,10 @@ class UserComboOption(UserOption):
         return value
 
 class UserArrayOption(UserOption):
-    def __init__(self, name, description, value, shlex_split=False, user_input=False, **kwargs):
+    def __init__(self, name, description, value, shlex_split=False, user_input=False, allow_dups=False, **kwargs):
         super().__init__(name, description, kwargs.get('choices', []), yielding=kwargs.get('yielding', None))
         self.shlex_split = shlex_split
+        self.allow_dups = allow_dups
         self.value = self.validate_value(value, user_input=user_input)
 
     def validate_value(self, value, user_input=True):
@@ -166,7 +170,7 @@ class UserArrayOption(UserOption):
         else:
             raise MesonException('"{0}" should be a string array, but it is not'.format(str(newvalue)))
 
-        if len(set(newvalue)) != len(newvalue):
+        if not self.allow_dups and len(set(newvalue)) != len(newvalue):
             msg = 'Duplicated values in array option "%s" is deprecated. ' \
                   'This will become a hard error in the future.' % (self.name)
             mlog.deprecation(msg)
@@ -180,6 +184,7 @@ class UserArrayOption(UserOption):
                     ', '.join(bad), ', '.join(self.choices)))
         return newvalue
 
+
 class UserFeatureOption(UserComboOption):
     static_choices = ['enabled', 'disabled', 'auto']
 
@@ -195,6 +200,72 @@ class UserFeatureOption(UserComboOption):
     def is_auto(self):
         return self.value == 'auto'
 
+
+def load_configs(filenames):
+    """Load native files."""
+    def gen():
+        for f in filenames:
+            f = os.path.expanduser(os.path.expandvars(f))
+            if os.path.exists(f):
+                yield f
+                continue
+            elif sys.platform != 'win32':
+                f = os.path.basename(f)
+                paths = [
+                    os.environ.get('XDG_DATA_HOME', os.path.expanduser('~/.local/share')),
+                ] + os.environ.get('XDG_DATA_DIRS', '/usr/local/share:/usr/share').split(':')
+                for path in paths:
+                    path_to_try = os.path.join(path, 'meson', 'native', f)
+                    if os.path.isfile(path_to_try):
+                        yield path_to_try
+                        break
+                else:
+                    raise MesonException('Cannot find specified native file: ' + f)
+                continue
+
+            raise MesonException('Cannot find specified native file: ' + f)
+
+    config = configparser.SafeConfigParser()
+    config.read(gen())
+    return config
+
+
+def _get_section(config, section):
+    if config.has_section(section):
+        final = {}
+        for k, v in config.items(section):
+            # Windows paths...
+            v = v.replace('\\', '\\\\')
+            try:
+                final[k] = ast.literal_eval(v)
+            except SyntaxError:
+                raise MesonException(
+                    'Malformed value in native file variable: {}'.format(v))
+        return final
+    return {}
+
+
+class ConfigData:
+
+    """Contains configuration information provided by the user for the build."""
+
+    def __init__(self, config=None):
+        if config:
+            self.binaries = _get_section(config, 'binaries')
+            # global is a keyword and globals is a builtin, rather than mangle it,
+            # use a similar word
+            self.universal = _get_section(config, 'globals')
+            self.subprojects = {s: _get_section(config, s) for s in config.sections()
+                                if s not in {'binaries', 'globals'}}
+        else:
+            self.binaries = {}
+            self.universal = {}
+            self.subprojects = {}
+
+    def get_binaries(self, name):
+        return self.binaries.get(name, None)
+
+
 # This class contains all data that must persist over multiple
 # invocations of Meson. It is roughly the same thing as
 # cmakecache.
@@ -211,6 +282,7 @@ class CoreData:
         }
         self.test_guid = str(uuid.uuid4()).upper()
         self.regen_guid = str(uuid.uuid4()).upper()
+        self.install_guid = str(uuid.uuid4()).upper()
         self.target_guids = {}
         self.version = version
         self.init_builtins()
@@ -220,12 +292,20 @@ class CoreData:
         self.base_options = {}
         self.external_preprocess_args = {} # CPPFLAGS only
         self.cross_file = self.__load_cross_file(options.cross_file)
-        self.wrap_mode = options.wrap_mode
         self.compilers = OrderedDict()
         self.cross_compilers = OrderedDict()
         self.deps = OrderedDict()
         # Only to print a warning if it changes between Meson invocations.
         self.pkgconf_envvar = os.environ.get('PKG_CONFIG_PATH', '')
+        self.config_files = self.__load_config_files(options.native_file)
+
+    @staticmethod
+    def __load_config_files(filenames):
+        if not filenames:
+            return []
+        filenames = [os.path.abspath(os.path.expanduser(os.path.expanduser(f)))
+                     for f in filenames]
+        return filenames
 
     @staticmethod
     def __load_cross_file(filename):
@@ -333,10 +413,12 @@ class CoreData:
                     'Default project to execute in Visual Studio',
                     '')
 
-
     def get_builtin_option(self, optname):
         if optname in self.builtins:
-            return self.builtins[optname].value
+            v = self.builtins[optname]
+            if optname == 'wrap_mode':
+                return WrapMode.from_string(v.value)
+            return v.value
         raise RuntimeError('Tried to get unknown builtin option %s.' % optname)
 
     def set_builtin_option(self, optname, value):
@@ -349,6 +431,51 @@ class CoreData:
             raise RuntimeError('Tried to set unknown builtin option %s.' % optname)
         self.builtins[optname].set_value(value)
 
+        # Make sure that buildtype matches other settings.
+        if optname == 'buildtype':
+            self.set_others_from_buildtype(value)
+        else:
+            self.set_buildtype_from_others()
+
+    def set_others_from_buildtype(self, value):
+        if value == 'plain':
+            opt = '0'
+            debug = False
+        elif value == 'debug':
+            opt = '0'
+            debug = True
+        elif value == 'debugoptimized':
+            opt = '2'
+            debug = True
+        elif value == 'release':
+            opt = '3'
+            debug = False
+        elif value == 'minsize':
+            opt = 's'
+            debug = True
+        else:
+            assert(value == 'custom')
+            return
+        self.builtins['optimization'].set_value(opt)
+        self.builtins['debug'].set_value(debug)
+
+    def set_buildtype_from_others(self):
+        opt = self.builtins['optimization'].value
+        debug = self.builtins['debug'].value
+        if opt == '0' and not debug:
+            mode = 'plain'
+        elif opt == '0' and debug:
+            mode = 'debug'
+        elif opt == '2' and debug:
+            mode = 'debugoptimized'
+        elif opt == '3' and not debug:
+            mode = 'release'
+        elif opt == 's' and debug:
+            mode = 'minsize'
+        else:
+            mode = 'custom'
+        self.builtins['buildtype'].set_value(mode)
+
     def validate_option_value(self, option_name, override_value):
         for opts in (self.builtins, self.base_options, self.compiler_options, self.user_options):
             if option_name in opts:
@@ -389,8 +516,7 @@ class CoreData:
             if k == 'prefix':
                 pass
             elif k in self.builtins:
-                tgt = self.builtins[k]
-                tgt.set_value(self.sanitize_dir_option_value(prefix, k, v))
+                self.set_builtin_option(k, v)
             elif k in self.backend_options:
                 tgt = self.backend_options[k]
                 tgt.set_value(v)
@@ -411,18 +537,78 @@ class CoreData:
             sub = 'In subproject {}: '.format(subproject) if subproject else ''
             mlog.warning('{}Unknown options: "{}"'.format(sub, unknown_options))
 
+class CmdLineFileParser(configparser.ConfigParser):
+    def __init__(self):
+        # We don't want ':' as key delimiter, otherwise it would break when
+        # storing subproject options like "subproject:option=value"
+        super().__init__(delimiters=['='])
+
+def get_cmd_line_file(build_dir):
+    return os.path.join(build_dir, 'meson-private', 'cmd_line.txt')
+
+def read_cmd_line_file(build_dir, options):
+    filename = get_cmd_line_file(build_dir)
+    config = CmdLineFileParser()
+    config.read(filename)
+
+    # Do a copy because config is not really a dict. options.cmd_line_options
+    # overrides values from the file.
+    d = dict(config['options'])
+    d.update(options.cmd_line_options)
+    options.cmd_line_options = d
+
+    properties = config['properties']
+    if options.cross_file is None:
+        options.cross_file = properties.get('cross_file', None)
+    if not options.native_file:
+        # This will be a string in the form: "['first', 'second', ...]", use
+        # literal_eval to get it into the list of strings.
+        options.native_file = ast.literal_eval(properties.get('native_file', '[]'))
+
+def write_cmd_line_file(build_dir, options):
+    filename = get_cmd_line_file(build_dir)
+    config = CmdLineFileParser()
+
+    properties = {}
+    if options.cross_file is not None:
+        properties['cross_file'] = options.cross_file
+    if options.native_file:
+        properties['native_file'] = options.native_file
+
+    config['options'] = options.cmd_line_options
+    config['properties'] = properties
+    with open(filename, 'w') as f:
+        config.write(f)
+
+def update_cmd_line_file(build_dir, options):
+    filename = get_cmd_line_file(build_dir)
+    config = CmdLineFileParser()
+    config.read(filename)
+    config['options'].update(options.cmd_line_options)
+    with open(filename, 'w') as f:
+        config.write(f)
+
+def major_versions_differ(v1, v2):
+    return v1.split('.')[0:2] != v2.split('.')[0:2]
+
 def load(build_dir):
     filename = os.path.join(build_dir, 'meson-private', 'coredata.dat')
     load_fail_msg = 'Coredata file {!r} is corrupted. Try with a fresh build tree.'.format(filename)
     try:
         with open(filename, 'rb') as f:
             obj = pickle.load(f)
-    except pickle.UnpicklingError:
+    except (pickle.UnpicklingError, EOFError):
         raise MesonException(load_fail_msg)
+    except AttributeError:
+        raise MesonException(
+            "Coredata file {!r} references functions or classes that don't "
+            "exist. This probably means that it was generated with an old "
+            "version of meson.".format(filename))
     if not isinstance(obj, CoreData):
         raise MesonException(load_fail_msg)
-    if obj.version != version:
-        raise MesonException('Build directory has been generated with Meson version %s, which is incompatible with current version %s.\nPlease delete this build directory AND create a new one.' %
+    if major_versions_differ(obj.version, version):
+        raise MesonException('Build directory has been generated with Meson version %s, '
+                             'which is incompatible with current version %s.\n' %
                              (obj.version, version))
     return obj
 
@@ -430,7 +616,7 @@ def save(obj, build_dir):
     filename = os.path.join(build_dir, 'meson-private', 'coredata.dat')
     prev_filename = filename + '.prev'
     tempfilename = filename + '~'
-    if obj.version != version:
+    if major_versions_differ(obj.version, version):
         raise MesonException('Fatal version mismatch corruption.')
     if os.path.exists(filename):
         import shutil
@@ -544,31 +730,37 @@ def parse_cmd_line_options(args):
             delattr(args, name)
 
 builtin_options = {
-    'buildtype':  [UserComboOption, 'Build type to use.', ['plain', 'debug', 'debugoptimized', 'release', 'minsize'], 'debug'],
-    'strip':      [UserBooleanOption, 'Strip targets on install.', False],
-    'unity':      [UserComboOption, 'Unity build.', ['on', 'off', 'subprojects'], 'off'],
-    'prefix':     [UserStringOption, 'Installation prefix.', default_prefix()],
-    'libdir':     [UserStringOption, 'Library directory.', default_libdir()],
-    'libexecdir': [UserStringOption, 'Library executable directory.', default_libexecdir()],
-    'bindir':     [UserStringOption, 'Executable directory.', 'bin'],
-    'sbindir':    [UserStringOption, 'System executable directory.', 'sbin'],
-    'includedir': [UserStringOption, 'Header file directory.', 'include'],
-    'datadir':    [UserStringOption, 'Data file directory.', 'share'],
-    'mandir':     [UserStringOption, 'Manual page directory.', 'share/man'],
-    'infodir':    [UserStringOption, 'Info page directory.', 'share/info'],
-    'localedir':  [UserStringOption, 'Locale data directory.', 'share/locale'],
-    'sysconfdir':      [UserStringOption, 'Sysconf data directory.', 'etc'],
-    'localstatedir':   [UserStringOption, 'Localstate data directory.', 'var'],
-    'sharedstatedir':  [UserStringOption, 'Architecture-independent data directory.', 'com'],
-    'werror':          [UserBooleanOption, 'Treat warnings as errors.', False],
-    'warning_level':   [UserComboOption, 'Compiler warning level to use.', ['1', '2', '3'], '1'],
-    'layout':          [UserComboOption, 'Build directory layout.', ['mirror', 'flat'], 'mirror'],
-    'default_library': [UserComboOption, 'Default library type.', ['shared', 'static', 'both'], 'shared'],
-    'backend':         [UserComboOption, 'Backend to use.', backendlist, 'ninja'],
-    'stdsplit':        [UserBooleanOption, 'Split stdout and stderr in test logs.', True],
-    'errorlogs':       [UserBooleanOption, "Whether to print the logs from failing tests.", True],
-    'install_umask':   [UserUmaskOption, 'Default umask to apply on permissions of installed files.', '022'],
-    'auto_features':   [UserFeatureOption, "Override value of all 'auto' features.", 'auto'],
+    'buildtype':  [UserComboOption, 'Build type to use', ['plain', 'debug', 'debugoptimized', 'release', 'minsize', 'custom'], 'debug'],
+    'strip':      [UserBooleanOption, 'Strip targets on install', False],
+    'unity':      [UserComboOption, 'Unity build', ['on', 'off', 'subprojects'], 'off'],
+    'prefix':     [UserStringOption, 'Installation prefix', default_prefix()],
+    'libdir':     [UserStringOption, 'Library directory', default_libdir()],
+    'libexecdir': [UserStringOption, 'Library executable directory', default_libexecdir()],
+    'bindir':     [UserStringOption, 'Executable directory', 'bin'],
+    'sbindir':    [UserStringOption, 'System executable directory', 'sbin'],
+    'includedir': [UserStringOption, 'Header file directory', 'include'],
+    'datadir':    [UserStringOption, 'Data file directory', 'share'],
+    'mandir':     [UserStringOption, 'Manual page directory', 'share/man'],
+    'infodir':    [UserStringOption, 'Info page directory', 'share/info'],
+    'localedir':  [UserStringOption, 'Locale data directory', 'share/locale'],
+    'sysconfdir':      [UserStringOption, 'Sysconf data directory', 'etc'],
+    'localstatedir':   [UserStringOption, 'Localstate data directory', 'var'],
+    'sharedstatedir':  [UserStringOption, 'Architecture-independent data directory', 'com'],
+    'werror':          [UserBooleanOption, 'Treat warnings as errors', False],
+    'warning_level':   [UserComboOption, 'Compiler warning level to use', ['1', '2', '3'], '1'],
+    'layout':          [UserComboOption, 'Build directory layout', ['mirror', 'flat'], 'mirror'],
+    'default_library': [UserComboOption, 'Default library type', ['shared', 'static', 'both'], 'shared'],
+    'backend':         [UserComboOption, 'Backend to use', backendlist, 'ninja'],
+    'stdsplit':        [UserBooleanOption, 'Split stdout and stderr in test logs', True],
+    'errorlogs':       [UserBooleanOption, "Whether to print the logs from failing tests", True],
+    'install_umask':   [UserUmaskOption, 'Default umask to apply on permissions of installed files', '022'],
+    'auto_features':   [UserFeatureOption, "Override value of all 'auto' features", 'auto'],
+    'optimization':    [UserComboOption, 'Optimization level', ['0', 'g', '1', '2', '3', 's'], '0'],
+    'debug':           [UserBooleanOption, 'Debug', True],
+    'wrap_mode':       [UserComboOption, 'Wrap mode', ['default',
+                                                       'nofallback',
+                                                       'nodownload',
+                                                       'forcefallback'], 'default'],
 }
 
 # Special prefix-dependent defaults for installation directories that reside in
