@@ -14,6 +14,7 @@
 
 import re
 import codecs
+import types
 from .mesonlib import MesonException
 from . import mlog
 
@@ -89,7 +90,9 @@ class Lexer:
     def __init__(self, code):
         self.code = code
         self.keywords = {'true', 'false', 'if', 'else', 'elif',
-                         'endif', 'and', 'or', 'not', 'foreach', 'endforeach'}
+                         'endif', 'and', 'or', 'not', 'foreach', 'endforeach',
+                         'in', 'continue', 'break'}
+        self.future_keywords = {'return'}
         self.token_specification = [
             # Need to be sorted longest to shortest.
             ('ignore', re.compile(r'[ \t]')),
@@ -196,6 +199,9 @@ This will become a hard error in a future Meson release.""", self.getline(line_s
                         if match_text in self.keywords:
                             tid = match_text
                         else:
+                            if match_text in self.future_keywords:
+                                mlog.warning("Identifier '{}' will become a reserved keyword in a future release. Please rename it.".format(match_text),
+                                             location=types.SimpleNamespace(subdir=subdir, lineno=lineno))
                             value = match_text
                     yield Token(tid, subdir, curline_start, curline, col, bytespan, value)
                     break
@@ -237,6 +243,12 @@ class StringNode(ElementaryNode):
     def __str__(self):
         return "String node: '%s' (%d, %d)." % (self.value, self.lineno, self.colno)
 
+class ContinueNode(ElementaryNode):
+    pass
+
+class BreakNode(ElementaryNode):
+    pass
+
 class ArrayNode:
     def __init__(self, args):
         self.subdir = args.subdir
@@ -431,7 +443,9 @@ comparison_map = {'equal': '==',
                   'lt': '<',
                   'le': '<=',
                   'gt': '>',
-                  'ge': '>='
+                  'ge': '>=',
+                  'in': 'in',
+                  'notin': 'not in',
                   }
 
 # Recursive descent parser for Meson's definition language.
@@ -538,6 +552,8 @@ class Parser:
         for nodename, operator_type in comparison_map.items():
             if self.accept(nodename):
                 return ComparisonNode(operator_type, left, self.e5())
+        if self.accept('not') and self.accept('in'):
+            return ComparisonNode('notin', left, self.e5())
         return left
 
     def e5(self):
@@ -749,6 +765,10 @@ class Parser:
             block = self.foreachblock()
             self.block_expect('endforeach', block_start)
             return block
+        if self.accept('continue'):
+            return ContinueNode(self.current)
+        if self.accept('break'):
+            return BreakNode(self.current)
         return self.statement()
 
     def codeblock(self):
